import md5
import sha
import struct
import binascii
import string
import sys
import otpdict

class OTP:
	def __init__(self, mdalg, seed, passphrase):
		if mdalg == 'otp-md5':
			self.digest = otpmd5
		elif mdalg == 'otp-sha1':
			self.digest = otpsha1
		else:
			raise "Invalid/Unsupported MD algorithm"

		self.mdalg = mdalg
		self.seed = string.lower(seed)
		self.passphrase = passphrase
			
	def getAlgorithm(self):
		return self.mdalg
	
	def getSeed(self):
		return self.seed

	def getRaw(self, sequence):
		if sequence < 1:
			raise "Sequence number less than 1"
		hashResult = self.digest(self.seed + self.passphrase)
		for i in range(sequence):
			hashResult = self.digest(hashResult)
		return hashResult
	
	def getHex(self, sequence):
		return binascii.b2a_hex(self.getRaw(sequence))
	
	def getWords(self, sequence):
		raw = self.getRaw(sequence)
		parity = 0;
		for c in raw:
			i = ord(c)
			parity = parity + i & 3
			parity = parity + (i >> 2) & 3
			parity = parity + (i >> 4) & 3
			parity = parity + (i >> 6) & 3
	
		(asLong,) = struct.unpack('>q', raw)
		# Beware the 66bit integer ;-)
		asLong = (asLong << 2) | parity
		words = []
		for i in (5,4,3,2,1,0):
			index = (asLong >> (i * 11)) & 2047
			words.append(otpdict.standard[index])

		return string.join(words, " ")


		
def otpmd5(string):
	md5digest = md5.new(string).digest()
	asInts = struct.unpack('>iiii', md5digest)
	return struct.pack('>ii', asInts[0] ^ asInts[2], asInts[1] ^ asInts[3])

def otpsha1(string): # Warning: Untested!
	shadigest = sha.new(string).digest()
	asInts = struct.unpack('>iiiii', shadigest)
	return struct.pack('>ii', asInts[0] ^ asInts[2] ^ asInts[4], asInts[1] ^ asInts[3])

def myHex(integer):
	hexDigits = '0123456789ABCDEF'
	retString = ''
	for i in range(8):
		retString = hexDigits[(integer >> (i*4)) & 0x0000000f] + retString
	if sys.byteorder == 'little':
		return retString[6]+retString[7]+retString[4]+retString[5]+retString[2]+retString[3]+retString[0]+retString[1]
	else:
		return retString[:4] + retString[4:8]
		
