#!/usr/bin/env python

import wxversion
wxversion.select("2.8")
import wx

from miknet.otp import otp
import string
import sys
import re
import socket
import select
import os
import os.path

if os.name == "posix":
	socketListener=1
	import thread
else:
	socketListener=0

class PyOTPFrame(wx.Frame):
	def __init__(self, *args, **kwargs):
		wx.Frame.__init__(self, *args, **kwargs)

		if socketListener:
			self.otpMutex = thread.allocate_lock()
			self.stopMutex = thread.allocate_lock()
			self._stopThread = 0

		self.Bind(wx.EVT_CLOSE, self.OnClose)
		
		menuBar = wx.MenuBar()
		fileMenu = wx.Menu()
		otpMenu = wx.Menu()

		quitItem = fileMenu.Append(wx.ID_EXIT, "&Quit")	
		self.Bind(wx.EVT_MENU, self.OnQuit, quitItem)

		newItem = otpMenu.Append(-1, "&New OTP\tCtrl-N")
		self.Bind(wx.EVT_MENU, self.newotp, newItem)

		reseedItem = otpMenu.Append(-1, "&Reseed OTP\tCtrl-R")
		self.Bind(wx.EVT_MENU, self.reseed, reseedItem)

		calcItem = otpMenu.Append(-1, "&Calculate\tF9")
		self.Bind(wx.EVT_MENU, self.calcotp, calcItem)

		fingerItem = otpMenu.Append(-1, "&Finger\tCtrl-F")
		self.Bind(wx.EVT_MENU, self.reFinger, fingerItem)

		menuBar.Append(fileMenu, "&File")
		menuBar.Append(otpMenu, "&Otp")

		self.SetMenuBar(menuBar)

		sizer = wx.BoxSizer(wx.HORIZONTAL)

		self.sequenceSpin = wx.SpinCtrl(self, -1, "")
		self.sequenceSpin.SetRange(1, 999)
		sizer.Add(self.sequenceSpin)
		
		calcButton = wx.Button(self, label="Calculate")
		self.Bind(wx.EVT_BUTTON, self.calcotp, calcButton)
		sizer.Add(calcButton)

		self.CreateStatusBar()
		self.SetSizer(sizer)
		self.SetAutoLayout(True)
		sizer.Fit(self)

		self.otp = None
		self.loadprefs()


		if not self.otp:
			self.Close(True)

		self.Show(True)
		if socketListener:
			thread.start_new_thread(self.listenForConnections, ())

	def OnQuit(self, event):
		self.Close(True)
	
	def OnClose(self, event):
		self.stopThread()
		if self.otp:
			self.saveprefs()
		self.Destroy()
	
	def reFinger(self, event=None):
		if self.fingerUser == "" or self.fingerServer == "":
			self.SetStatusText("No finger info available")
			return
		try:
			alg, sequence, seed = FingerClient(self.fingerUser, self.fingerServer, self.fingerPort).doFinger()
			if self.otp.seed != seed:
				self.SetStatusText("Seed changed, not setting sequence")
				return
			self.SetStatusText("Finger Successful")
			self.sequenceSpin.SetValue(int(sequence))
		except Exception, e:
			print str(e)
			
	def newotp(self, Event = None, seed = None, algorithm = None, sequence = None, fingerUser = None, fingerServer = None, fingerPort = 79):
		dialog = OTPDialog(self, -1, seed, algorithm, sequence, fingerUser, fingerServer, fingerPort)
		if dialog.ShowModal() == wx.ID_OK:
			if socketListener:
				self.otpMutex.acquire()
			self.otp = otp.OTP("otp-md5", dialog.seedText.GetValue(), dialog.passphraseText.GetValue())
			self.sequenceSpin.SetValue(dialog.sequenceSpin.GetValue())
			if socketListener:
				self.otpMutex.release()
			self.fingerUser = dialog.fingerUserText.GetValue()
			self.fingerServer = dialog.fingerServerText.GetValue()
			try:
				self.fingerPort = int(dialog.fingerPortText.GetValue())
			except:
				self.fingerPort = 79
	
	def stopThread(self):
		if socketListener:
			self.stopMutex.acquire()
			self._stopThread = 1
			self.stopMutex.release()

	def _isStopped(self):
		self.stopMutex.acquire()
		_stopThread = self._stopThread
		self.stopMutex.release()
		return _stopThread

	def listenForConnections(self):
		s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
		socketFile = os.path.join(os.environ["HOME"], ".mikotp.sock")
		try: os.unlink(socketFile)
		except: pass
		s.bind(socketFile)
		s.listen(1)
		while not self._isStopped():
			i, o, e = select.select((s,), (), (), 1)
			if len(i) > 0:
				try:
					conn, addr = s.accept()
					conn.send(self._doCalc())
					conn.shutdown(socket.SHUT_RDWR)
				except Exception, e:
					print "Error talking on socket: %s" % str(e)
		s.close()
	
	def loadprefs(self):
		algorithm = None
		seed = None
		sequence = None
		fingerUser = None
		fingerServer = None
		fingerPort = 79
		
		try:
			fh = None
			try:
				fh = open(os.path.join(os.environ["HOME"], ".mikotp.properties"), "r")
			except:
				fh = open("jotp.properties", "r")

			buf = fh.readline()
			while len(buf) > 0:
				buf = buf.strip()
				splitString = buf.split("=", 2)
				if len(splitString) == 2:
					key = splitString[0].strip()
					value = splitString[1].strip()
				else:
					buf = fh.readline()
					continue
				
				if key == "jotp.algorithm":
					if value == "MD4":
						algorithm = "otp-md4"
					elif value == "MD5":
						algorithm = "otp-md5"
					else:
						raise "Invalid OTP algorithm"
				elif key == "jotp.seed":
					seed = value
				elif key == "jotp.sequence":
					sequence = int(value)
				elif key == "mikotp.fingerUser":
					fingerUser = value
				elif key == "mikotp.fingerServer":
					fingerServer = value
				elif key == "mikotp.fingerPort":
					fingerPort = int(value)
				buf = fh.readline()
		except:
			print "Got an exception while reading prefs:", sys.exc_info()[0]
			pass

		self.newotp(None, seed, algorithm, sequence, fingerUser, fingerServer, fingerPort)

	def saveprefs(self):
		if os.environ.has_key("HOME"):
			filename = os.path.join(os.environ["HOME"], ".mikotp.properties")
		else:
			filename = "jotp.properties"

		try:
			fh = open(filename, "w")
		except:
			print "Warning: unable to save preferences to jotp.properties"
			return

		alg = self.otp.getAlgorithm()
		if alg == "otp-md4":
			fh.write("jotp.algorithm=MD4\n")
		else:
			fh.write("jotp.algorithm=MD5\n")

		fh.write("jotp.seed=" + self.otp.getSeed() + "\n")
		fh.write("jotp.sequence=" + str(self.sequenceSpin.GetValue()) + "\n")

		if self.fingerUser != "":
			fh.write("mikotp.fingerUser=" + self.fingerUser + "\n")
		if self.fingerServer != "":
			fh.write("mikotp.fingerServer=" + self.fingerServer + "\n")
		if self.fingerPort != 79:
			fh.write("mikotp.fingerPort=" + str(self.fingerPort) + "\n")

		fh.close()

	def _doCalc(self):
		if socketListener:
			self.otpMutex.acquire()
		sequence = self.sequenceSpin.GetValue()
		result = self.otp.getHex(sequence)
		self.sequenceSpin.SetValue(sequence-1)
		self.SetStatusText("OTP Sequence %d Calculated" % sequence)
		if socketListener:
			self.otpMutex.release()
		return result

	def calcotp(self, event = None):
		result = self._doCalc()
		clipdata = wx.TextDataObject()
		clipdata.SetText(result + "\n")
		wx.TheClipboard.Open()
		wx.TheClipboard.SetData(clipdata)
		wx.TheClipboard.Close()
		
	def reseed(self, Event = None):
		if socketListener:
			self.otpMutex.acquire()
		otpBackup = self.otp
		seqBackup = self.sequenceSpin.GetValue()
		oldotp = self.otp.getHex(seqBackup)
		match = re.match("([a-zA-Z]+)([0-9]+)", self.otp.getSeed())
		if match is None:
			self.SetStatusText("Warning: Unable to parse seed")
			if socketListener:
				self.otpMutex.release()
			return
		seedchars = match.group(1)
		seednumber = int(match.group(2)) + 1
		self.otp = otp.OTP("otp-md5", seedchars + str(seednumber), self.otp.passphrase)
		reinitstring = "init-hex:" + oldotp + ":md5 500 " + self.otp.getSeed() + ":" + self.otp.getHex(500) + "\n"
		self.sequenceSpin.SetValue(499)
		if socketListener:
			self.otpMutex.release()
		clipdata = wx.TextDataObject()
		clipdata.SetText(reinitstring)
		wx.TheClipboard.Open()
		wx.TheClipboard.SetData(clipdata)
		wx.TheClipboard.Close()
		result = wx.MessageDialog(self, "Was Re-Init Successful?").ShowModal()
		if result != wx.ID_OK:
			self.otp = otpBackup
			self.sequenceSpin.SetValue(seqBackup)


class OTPDialog(wx.Dialog):
	def __init__(self, parent, id,  seed = None, algorithm = None, sequence = None, fingerUser = None, fingerServer = None, fingerPort = 79):
		wx.Dialog.__init__(self, parent, id, "Create OTP")

		self.otp = None

		fingerUserLabel = wx.StaticText(self, label="Finger User:")
		if fingerUser is None:
			fingerUser = ""
		self.fingerUserText = wx.TextCtrl(self, value=fingerUser)

		fingerServerLabel = wx.StaticText(self, label="Finger Server:")
		if fingerServer is None:
			fingerServer = ""
		self.fingerServerText = wx.TextCtrl(self, value=fingerServer, size=(200, -1))

		fingerPortLabel = wx.StaticText(self, label="Finger Port:")
		self.fingerPortText = wx.TextCtrl(self, value=str(fingerPort))

		fingerButtonLabel = wx.StaticText(self, label="Finger")
		fingerButton = wx.Button(self, label="Grab info")
		self.Bind(wx.EVT_BUTTON, self.doFinger, fingerButton)

		if sequence:
			sequence = sequence
		else:
			sequence = 499

		seedLabel = wx.StaticText(self, label="Seed:")
		if seed:
			seedText = str(seed)
		else:
			seedText = ""
		self.seedText = wx.TextCtrl(self, value=seedText)

		passphraseLabel = wx.StaticText(self, label="Passphrase:")
		self.passphraseText = wx.TextCtrl(self, size=(200, -1), style=wx.TE_PASSWORD)

		sequenceLabel = wx.StaticText(self, label="Sequence:")
		self.sequenceSpin = wx.SpinCtrl(self)
		self.sequenceSpin.SetRange(1, 999)
		self.sequenceSpin.SetValue(sequence)

		OkButton = wx.Button(self, wx.ID_OK, "OK")
		OkButton.SetDefault()
		CancelButton = wx.Button(self, wx.ID_CANCEL, "Cancel")
		
		sizer = wx.FlexGridSizer(8, 2, 2, 2) # rows, cols, hgap, vgap
		sizer.Add(fingerUserLabel)
		sizer.Add(self.fingerUserText)
		sizer.Add(fingerServerLabel)
		sizer.Add(self.fingerServerText)
		sizer.Add(fingerPortLabel)
		sizer.Add(self.fingerPortText)
		sizer.Add(fingerButtonLabel)
		sizer.Add(fingerButton)
		sizer.Add(seedLabel)
		sizer.Add(self.seedText)
		sizer.Add(passphraseLabel)
		sizer.Add(self.passphraseText)
		sizer.Add(sequenceLabel)
		sizer.Add(self.sequenceSpin)
		sizer.Add(OkButton, 0, wx.ALIGN_CENTER|wx.ALL, 5)
		sizer.Add(CancelButton, 0, wx.ALIGN_CENTER|wx.ALL, 5)
		
		if seed:
			self.passphraseText.SetFocus()
		self.SetSizer(sizer)
		self.SetAutoLayout(True)
		sizer.Fit(self)

	def doFinger(self, Event=None):
		user = self.fingerUserText.GetValue()
		server = self.fingerServerText.GetValue()
		port = 79
		try: port = int(self.fingerPortText.GetValue())
		except: pass

		def ErrorMessage(string):
			wx.MessageDialog(self, string, style=wx.OK).ShowModal()

		if user == "":
			ErrorMessage("Finger user not specified")
			self.fingerUserText.SetFocus()
			return
		elif server == "":
			ErrorMessage("Finger server not specified")
			self.fingerServerText.SetFocus()
			return

		try:
			alg, sequence, seed = FingerClient(user, server, port).doFinger()
			self.sequenceSpin.SetValue(int(sequence))
			self.seedText.SetValue(seed)
			self.passphraseText.SetFocus()
		except socket.error, e:
			ErrorMessage("Error connecting to server: %s" % str(e))
			return
		except FingerError, e:
			ErrorMessage(str(e))
			return

class FingerError(Exception):
	pass

class FingerClient:
	SELECT_TIMEOUT = 5
	parseChallenge = re.compile("otp-([a-z0-9]+) ([0-9]+) ([a-z0-9]+) ext").match

	def __init__(self, user, server, port = 79):
		self.user = user
		self.port = port
		self.server = server

	def doFinger(self):
		s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		s.connect((self.server, self.port))
		s.send(self.user + "\r\n")
		irdy, ordy, erdy = select.select((s,), tuple(), tuple(), self.SELECT_TIMEOUT)
		if len(irdy) == 0:
			raise FingerError("Timeout waiting for server to respond")
		challenge = s.recv(4096)

		matchObj = self.parseChallenge(challenge)
		if matchObj is None:
			raise FingerError("Invalid response from finger server")
		return matchObj.groups()

app = wx.App()
frame = PyOTPFrame(None, -1, "MikOTP")
app.MainLoop()

